
class Status(object):
    def __init__(self, maxval):
        pass
    
    def update(self, num):
        pass
    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass
